import efemarai.metamorph.base_metamorphs as base

# Each import is loading the corresponding operators in base.METAMORPHS
import efemarai.metamorph.text.text_metamorphs
import efemarai.metamorph.vision.vision_metamorphs

METAMORPHS = base.METAMORPHS
